const path = require('path');
const fse = require('fs-extra');
const assert = require('assert');
const deepmerge = require('deepmerge');

const log = require('../utils/log');
const getPkgData = require('../config/getPackageJson');
const getDefaultWebpackConfig = require('../config/getDefaultWebpackConfig');
const processEntry = require('../config/processEntry');
const PluginAPI = require('./Plugin');
const defaultConfig = require('../config/default.config');

module.exports = class Context {
  constructor({ command = '', rootDir = process.cwd(), args = {} }) {
    this.command = command;
    this.commandArgs = args;
    this.rootDir = rootDir;
    this.pkg = getPkgData(this.rootDir);
    this.userConfig = this.getUserConfig();
    this.plugins = this.getPlugins();
    this.chainWebpackFns = [];
    this.eventHooks = {};

    this.reRun = this.reRun.bind(this);
    this.applyHook = this.applyHook.bind(this);
  }

  getUserConfig() {
    const { config } = this.commandArgs;
    let zendConfigPath = '';
    if (config) {
      zendConfigPath = path.isAbsolute(config)
        ? config
        : path.resolve(this.rootDir, config);
    } else {
      zendConfigPath = path.resolve(this.rootDir, 'zend.config.js');
    }
    let userConfig = {
      // 默认@为src路径别名
      alias: {
        '@': path.resolve(this.rootDir + '/src')
      },
    };
    if (fse.existsSync(zendConfigPath)) {
      try {
        userConfig = require(zendConfigPath);
      } catch (err) {
        log.error(`Fail to load config file ${zendConfigPath}, use default config instead`);
        console.error(err);
        process.exit(1);
      }
    } else if (this.pkg.buildConfig) {
      log.warn('You should migrate config into zend.config.js');
    }
    return deepmerge(defaultConfig, userConfig);
  }

  getPlugins() {
    const builtInPlugins = [
      '../plugins/userConfig',
      '../plugins/cliOptions',
      '../plugins/devWatcher',
    ];
    return builtInPlugins.map((pluginPath) => require(pluginPath))
      .concat(this.userConfig.plugins || []);
  }

  async runPlugins() {
    for (let i = 0; i < this.plugins.length; i++) {
      let pluginInfo = this.plugins[i];
      if (!Array.isArray(pluginInfo)) {
        pluginInfo = [pluginInfo];
      }
      const [plugin, options] = pluginInfo;
      const pluginName = typeof plugin === 'string' ? plugin : '';
      try {
        const pluginFunc = typeof plugin === 'string'
          ? require(require.resolve(plugin, { paths: [this.rootDir] }))
          : plugin;
        assert(typeof pluginFunc === 'function', 'plugin must export a function');
        await pluginFunc(new PluginAPI(this, pluginName), options);
      } catch (err) {
        log.error(`Fail to load plugin ${pluginName}`);
        console.error(err);
        process.exit(1);
      }
    }
  }

  async applyHook(key, opts = {}) {
    const hooks = this.eventHooks[key] || [];
    const results = [];
    hooks.forEach((fn) => {
      try {
        results.push(fn(opts));
      } catch (e) {
        log.error(e);
        log.error(`Fail to excute hook ${key}`);
      }
    });
    await Promise.all(results);
  }

  getWebpackConfig() {
    const config = getDefaultWebpackConfig(this.command, this.userConfig);

    this.chainWebpackFns.forEach(({ fn, pluginName }) => {
      try {
        fn(config);
      } catch (err) {
        log.error(`Fail to exec plugin chainWebpack ${pluginName}`);
        throw (err);
      }
    });

    if (this.userConfig.chainWebpack) {
      try {
        this.userConfig.chainWebpack(config, {
          command: this.command,
          commandArgs: this.commandArgs,
        });
      } catch (err) {
        log.error('Fail to exec userConfig.chainWebpack');
        throw (err);
      }
    }
    processEntry(config, {
      hotDev: this.command === 'dev' && !this.commandArgs.disabledReload,
      polyfill: this.userConfig.injectBabel === 'polyfill',
    });
    return config.toConfig();
  }

  async reRun() {
    this.chainWebpackFns = [];
    this.eventHooks = {};
    this.run();
  }

  async run() {
    await this.runPlugins();
    this.webpackConfig = this.getWebpackConfig();
    const command = require(`../commands/${this.command}`);
    await command(this);
  }
};
